#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <utime.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#define MYNAME "mailcount"
#define VERSION "0.05"

const char *usage=
	"\n"
	MYNAME ": prints how many messages you have\n"
	"\n"
	"Usage: " MYNAME " [-h | -d] mailbox [mailbox mailbox ...]\n"
	"       " MYNAME " --help\n"
	"       " MYNAME " --version\n"
	"\n"
	"For example if you have $MAIL set, try ``" MYNAME " $MAIL'',\n"
	"otherwise try ``" MYNAME " /var/mail/your_login_name''.\n"
	"\n"
	"Written by Egmont Koblinger <egmont@fazekas.hu>\n"
	"\n";

const char *version=
	MYNAME " version " VERSION "\n";

enum { notspecial, empty, from_space, from_colon, status };


int linetype (const char *line)
{
	if (line[0] == '\n') return empty;
	if (!strncmp(line, "From ", 5)) return from_space;
	if (!strncmp(line, "From: ", 6)) return from_colon;
	if (!strncmp(line, "Status: ", 8)) return status;
	return notspecial;
}

int mailcount (const char *file, int *all, int *new, int *unread)
{
	struct stat mailboxstat;
	struct timeval tv[2];
	int fd;
	char *p, *end;
	size_t l;
	int internal, s, r, head, t;
	
	fd=open(file, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, MYNAME ": unable to open %s: %s\n",
		    file, strerror(errno));
		return 1;
	}
	if (fstat(fd, &mailboxstat)) {
		fprintf(stderr, MYNAME ": unable to fstat %s: %s\n",
		    file, strerror(errno));
		close(fd);
		return 1;
	}
	if (!S_ISREG(mailboxstat.st_mode)) {
		fprintf(stderr, MYNAME ": %s is not a regular file\n", file);
		close(fd);
		return 1;
	}
	l=mailboxstat.st_size;
	p=mmap(NULL, l, PROT_READ, MAP_SHARED, fd, 0);
	if (p == MAP_FAILED) {
		fprintf(stderr, MYNAME ": unable to mmap %s: %s\n",
		    file, strerror(errno));
		close(fd);
		return 1;
	}
	end = p+l;

	if (l==0) {
		munmap(p, l);
		close(fd);
		return 0;
	}
	if (p[l-1] != '\n') {
		fprintf(stderr, MYNAME ": invalid mailbox %s: doesn't end"
		    " with newline\n", file);
		munmap(p, l);
		close(fd);
		return 1;
	}
	if (linetype(p) != from_space) {
		fprintf(stderr, MYNAME ": invalid mailbox %s: doesn't start"
		    " with From\n", file);
		munmap(p, l);
		close(fd);
		return 1;
	}

	head=0;
	internal=s=r=0;

	while (p != end) {
		t=linetype(p);
		if (head) {
			if (t==from_colon) {
				if (!strncmp(p+6, "Mail System Internal Data", 25))
					internal=1;
				}
			if (t==status) {
				s=1;
				p+=8;
				while (*p != '\n') if (*p++ == 'R') r=1;
				}
			if (t==empty) {
				head=0;
				if (!internal) {
					(*all)++;
					if (!s) (*new)++;
					if (s && !r) (*unread)++;
				}
				internal=s=r=0;
			}
		}
		else {
			if (t==from_space) {
				head=1;
			}
		}
		while (*p++ != '\n');
	}

	if (head) {
		fprintf(stderr, MYNAME ": invalid mailbox %s: ends in"
		    " message header\n", file);
		munmap(p, l);
		close(fd);
		return 1;
	}
	munmap(p, l);
	tv[0].tv_sec = mailboxstat.st_atim.tv_sec;
	tv[0].tv_usec = mailboxstat.st_atim.tv_nsec / 1000;
	if (fstat(fd, &mailboxstat)) {
		fprintf(stderr, MYNAME ": unable to fstat %s: %s\n",
		    file, strerror(errno));
		close(fd);
		return 1;
	}
	tv[1].tv_sec = mailboxstat.st_mtim.tv_sec;
	tv[1].tv_usec = mailboxstat.st_mtim.tv_nsec / 1000;
	close(fd);
	utimes(file, tv);
	return 0;
}

int main (int argc, char *argv[])
{
	char res[100];
	int args;
	int all, new, unread, rd;
	int opt=0;

	if (argc < 2) {
		fprintf(stderr, "%s", usage);
		return 1;
	}
	if (!strcmp(argv[1], "--help")) {
		printf("%s", usage);
		return 0;
	}
	if (!strcmp(argv[1], "--version")) {
		printf("%s", version);
		return 0;
	}
	while (argv[1] != NULL && (!strcmp(argv[1], "-h") || !strcmp(argv[1], "-H") || !strcmp(argv[1], "-d") || !strcmp(argv[1], "-D"))) {
	    if ((!strcmp(argv[1], "-h") || !strcmp(argv[1], "-H")) && (opt == 0)) {
		opt=1;   /* Hungarian */
	    }
	    if ((!strcmp(argv[1], "-d") || !strcmp(argv[1], "-D")) && (opt == 0)) {
		opt=2;   /* Digits only */
	    }
		argv++;
		argc--;
	}
	if (argc < 2) {
		fprintf(stderr, "%s", usage);
		return 1;
	}

	all=new=unread=0;
	args=argc-1;
	while (args) {
		if (mailcount(argv[args], &all, &new, &unread)) return 1;
		args--;
	}
	rd=all-new-unread;

	res[0]=0;
	if (opt == 1) {
		if (new) sprintf(res+strlen(res), "%d j", new);
		if (new && unread && rd) sprintf(res+strlen(res), ", ");
		else if (new && (unread || rd)) sprintf(res+strlen(res), " s ");
		if (unread) sprintf(res+strlen(res), "%d olvasatlan", unread);
		if (unread && rd) sprintf(res+strlen(res), " s ");
		if (rd) sprintf(res+strlen(res), "%d olvasott", rd);
		if (new || unread || rd) sprintf(res+strlen(res), " leveled van.");
		else sprintf(res+strlen(res), "Nincs leveled.");
	} else if (opt == 2) {
		sprintf(res, "%d %d %d", new, unread, rd);
	} else {
		sprintf(res, "You have ");
		if (new) sprintf(res+strlen(res), "%d new message%s, ", new, new>1?"s":"");
		if (unread) sprintf(res+strlen(res), "%d unread message%s, ", unread, unread>1?"s":"");
		if (rd) sprintf(res+strlen(res), "%d read message%s, ", rd, rd>1?"s":"");
		if (new || unread || rd) sprintf(res+strlen(res)-2, ".");
		else sprintf(res+strlen(res), "no messages.");
	}
	printf("%s%s", res, opt>1?"":"\n");
	return 0;
}

